import pytest
import socket
import time
import anyio

from contextlib import closing
from aioserf import connection

from async_generator import asynccontextmanager


def extract_addr(rpc, ip_address, address_family=socket.AF_INET6):
    packed_ip_format = socket.inet_pton(address_family, ip_address)
    r = rpc._decode_addr_key({b'Addr': packed_ip_format})
    return r[b'Addr'].decode('utf-8')

@asynccontextmanager
async def rpc_connect(**kw):
    async with anyio.create_task_group() as tg:
        conn = connection.SerfConnection(tg, **kw)
        async with conn._connected():
            yield conn

class TestSerfConnection(object):
    """
    Tests for the Serf RPC communication object.
    """

    @pytest.mark.anyio
    async def test_has_a_default_host_port_and_timeout(self):
        async with rpc_connect() as rpc:
            assert rpc.host == 'localhost'
            assert rpc.port == 7373

    @pytest.mark.anyio
    async def test_allows_passing_host_port_and_timeout(self):
        async with anyio.create_task_group() as tg:
            rpc = connection.SerfConnection(tg, host='foo', port=455)
            assert rpc.host == 'foo'
            assert rpc.port == 455

    @pytest.mark.anyio
    async def test_representation(self):
        async with rpc_connect() as rpc:
            assert str(rpc) == \
                '<SerfConnection counter=0 host=localhost port=7373>'

    @pytest.mark.anyio
    async def test_connection_to_bad_socket_throws_exception(self):
        with pytest.raises(
                    connection.SerfConnectionError
                ) as exceptionInfo:
            async with rpc_connect(port=40000) as rpc:
                pass
        assert isinstance(exceptionInfo.value.__cause__, ConnectionRefusedError)

    @pytest.mark.anyio
    async def test_handshake_to_serf_agent(self):
        async with rpc_connect() as rpc:
            assert (await rpc.handshake()).head == {b'Seq': 0, b'Error': b''}

    @pytest.mark.anyio
    async def test_call_throws_exception_if_socket_none(self):
        async with rpc_connect() as rpc:
            with pytest.raises(connection.SerfError) as exceptionInfo:
                await rpc.call('members')
            assert 'handshake' in str(exceptionInfo).lower()

    @pytest.mark.anyio
    async def test_handshake_and_call_increments_counter(self):
        async with rpc_connect() as rpc:
            assert 'counter=0' in str(rpc)
            await rpc.handshake()
            assert 'counter=1' in str(rpc)
            await rpc.call('event',
                    {"Name": "foo", "Payload": "test payload", "Coalesce": True},
                    expect_body=False)
            assert 'counter=2' in str(rpc)

    @pytest.mark.anyio
    async def test_msgpack_object_stream_decode(self):
        async with rpc_connect() as rpc:
            await rpc.handshake()
            result = await rpc.call('members')
            assert result.head == {b'Error': b'', b'Seq': 1}
            assert b'Members' in result.body.keys()

    @pytest.mark.anyio
    async def test_small_socket_recv_size(self):
        async with rpc_connect() as rpc:
            # Read a paltry 7 bytes at a time, intended to stress the buffered
            # socket reading and msgpack message handling logic.
            await rpc.handshake()
            rpc._socket_recv_size = 7
            result = await rpc.call('members')
            assert result.head == {b'Error': b'', b'Seq': 1}
            assert b'Members' in result.body.keys()

    @pytest.mark.anyio
    @pytest.mark.xfail
    async def test_rpc_timeout(self):
        async with rpc_connect() as rpc:
            # Avoid delaying the test too much.
            rpc.timeout = 0.1
            await rpc.handshake()
            with pytest.raises(connection.SerfTimeout):
                # Incorrectly set expect_body to True for an event RPC,
                # which will wait around for a body it'll never get,
                # which should cause a SerfTimeout exception.
                rpc.call('event',
                        {"Name": "foo", "Payload": "test payload",
                        "Coalesce": True},
                        expect_body=True)

    @pytest.mark.anyio
    @pytest.mark.xfail
    async def test_connection_closed(self):
        async with rpc_connect() as rpc:
            await rpc.handshake()

            await rpc._socket.close()

            # maybe TODO: raises OSError
            with pytest.raises(connection.SerfConnectionError):
                await rpc.handshake()

    @pytest.mark.anyio
    async def test_decode_addr_key_ipv6(self):
        async with rpc_connect() as rpc:
            ip_address = '2001:a:b:c:1:2:3:4'
            assert extract_addr(rpc, ip_address) == ip_address

    @pytest.mark.anyio
    async def test_decode_addr_key_ipv4_mapped_ipv6(self):
        async with rpc_connect() as rpc:
            assert extract_addr(rpc, '::ffff:192.168.0.1') == '192.168.0.1'

    @pytest.mark.anyio
    async def test_decode_addr_key_ipv4(self):
        async with rpc_connect() as rpc:
            ip_address = '192.168.0.1'
            assert extract_addr(rpc, ip_address, socket.AF_INET) == ip_address

    @pytest.mark.anyio
    async def test_close(self):
        async with rpc_connect() as rpc:
            assert rpc._socket is not None
        assert rpc._socket is None



import socket
import sys
import msgpack
import resource
from async_generator import asynccontextmanager
import anyio
from anyio.exceptions import ClosedResourceError
import outcome

from .result import SerfResult
from .util import ValueEvent
from .exceptions import SerfError, SerfConnectionError, SerfClosedError

from logging import getLogger
logger = getLogger(__name__)

_conn_id = 0

class StreamReply:
    _running = False
    send_stop = True

    def __init__(self, conn, command, params, seq, expect_body):
        self._conn = conn
        self._command = command
        self._params = params
        self.seq = seq
        self.q = anyio.create_queue(10000)
        self.expect_body = -expect_body
    async def set(self, value):
        await self.q.put(outcome.Value(value))
    async def set_error(self, err):
        await self.q.put(outcome.Error(err))
    async def get(self):
        res = await self.q.get()
        if res is not None:
            res = res.unwrap()
        return res
    def __aiter__(self):
        if not self._running:
            pass # raise RuntimeError("You need to wrap this in an 'async with'")
        return self
    async def __anext__(self):
        if not self._running:
            raise StopAsyncIteration
        res = await self.q.get()
        if res is None:
            raise StopAsyncIteration
        return res.unwrap()
    async def __aenter__(self):
        reply = await self._conn._call(self._command, self._params, _reply=self)
        res = await reply.get()
        if res is not None:
            self.head = res.head
            self._running = True
        return self
    async def __aexit__(self, *exc):
        self._running = False
        hdl = self._conn._handlers
        if self.send_stop:
            async with anyio.open_cancel_scope(shield=True):
                await self._conn.call("stop", params={b'Stop':self.seq}, expect_body=False)
                if hdl is not None:
                    # TODO remember this for a while?
                    del hdl[self.seq]
    async def cancel(self):
        await self.q.put(None)

class SerfConnection(object):
    """
    Manages RPC communication to and from a Serf agent.
    """

    # Read from the RPC socket in blocks of this many bytes.
    # (Typically 4k)
    _socket_recv_size = resource.getpagesize()
    _conn_id = 0

    def __init__(self, tg, host='localhost', port=7373):
        global _conn_id
        _conn_id += 1
        self._conn_id = _conn_id
        self.tg = tg
        self.host = host
        self.port = port
        self._socket = None
        self._seq = 0
        self._handlers = {}

    # handler protocol: incoming messages are passed in using `.set`.
    # If .expect_body is True then the reader will add a body to the
    # request. If it's -1 then the first reply is the body-less OK (we
    # hope) and only subsequent replies will have a body.

    def __repr__(self):
        return "<%(class)s counter=%(c)s host=%(h)s port=%(p)s>" \
            % {'class': self.__class__.__name__,
               'c': self._seq,
               'h': self.host,
               'p': self.port}

    def stream(self, command, params=None, expect_body=True):
        """
        Sends the provided command to Serf for evaluation, with
        any parameters as the message body. Expect a streamed reply.

        Returns a StreamReply object which affords an async context manager
        plus async iterator, which will return replies.
        """
        return StreamReply(self, command, params, self._counter, expect_body)


    async def _call(self, command, params=None, expect_body=True, *, _reply=None):
        """
        Sends the provided command to Serf for evaluation, with
        any parameters as the message body.

        Returns the reply, if any.
        """
        class SingleReply(ValueEvent):
            def __init__(slf, seq, expect_body):
                super().__init__()
                slf.seq = seq
                slf.expect_body = expect_body
            async def set(slf,val):
                if self._handlers is not None:
                    del self._handlers[slf.seq]
                await super().set(val)
            async def set_error(slf,err):
                if self._handlers is not None:
                    del self._handlers[slf.seq]
                await super().set_error(err)

        if self._socket is None:
            return

        if _reply is None:
            seq = self._counter
            _reply = SingleReply(seq, expect_body)
        else:
            seq = _reply.seq
        if self._handlers is not None:
            self._handlers[seq] = _reply
        else:
            _reply = None

        if params:
            logger.debug("%d:Send %s:%s =%s", self._conn_id,seq,command, repr(params))
        else:
            logger.debug("%d:Send %s:%s", self._conn_id,seq,command)
        msg = msgpack.packb({"Seq": seq, "Command": command})
        if params is not None:
            msg += msgpack.packb(params)

        await self._socket.send_all(msg)

        return _reply

    async def call(self, command, params=None, expect_body=True):
        res = await self._call(command, params, expect_body=expect_body)
        if res is None:
            return res
        return await res.get()

    async def _handle_msg(self, msg):
        """Handle an incoming message"""
        if self._handlers is None:
            logger.warn("Message without handlers:%s", msg)
            return
        try:
            seq = msg.head[b'Seq']
        except KeyError:
            raise RuntimeError("Reader got out of sync: "+str(msg))
        try:
            hdl = self._handlers[seq]
        except KeyError:
            logger.warn("Spurious message %s: %s", seq, msg)
            return

        if msg.body is None and hdl.expect_body > 0 and (hdl.expect_body > 1 or not msg.head[b'Error']):
            return True
        # Do this here because stream replies might arrive immediately
        # i.e. before the queue listener gets around to us
        if hdl.expect_body < 0:
            hdl.expect_body = -hdl.expect_body
        if msg.head[b'Error']:
            await hdl.set_error(SerfError(msg))
            await anyio.sleep(0.01)
        else:
            await hdl.set(msg)
        return False


    async def _reader(self, scope):
        """Main loop for reading"""
        unpacker = msgpack.Unpacker(object_hook=self._decode_addr_key)
        cur_msg = None

        async with anyio.open_cancel_scope(shield=True) as s:
            await scope.set(s)

            try:
                while self._socket is not None:
                    if cur_msg is not None:
                        logger.debug("%d:wait for body", self._conn_id)
                    try:
                        buf = await self._socket.receive_some(self._socket_recv_size)
                    except ClosedResourceError:
                        return # closed by us
                    if len(buf) == 0:  # Connection was closed.
                        raise SerfClosedError("Connection closed by peer")
                    unpacker.feed(buf)

                    for msg in unpacker:
                        if cur_msg is not None:
                            logger.debug("%d  Body=%s",self._conn_id,msg)
                            cur_msg.body = msg
                            await self._handle_msg(cur_msg)
                            cur_msg = None
                        else:
                            logger.debug("%d:Recv =%s",self._conn_id, msg)
                            msg = SerfResult(msg)
                            if await self._handle_msg(msg):
                                cur_msg = msg
            finally:
                hdl, self._handlers = self._handlers, None
                async with anyio.open_cancel_scope(shield=True):
                    for m in hdl.values():
                        await m.cancel()

    async def handshake(self):
        """
        Sets up the connection with the Serf agent and does the
        initial handshake.
        """
        return await self.call('handshake', {"Version": 1}, expect_body=False)

    async def auth(self, auth_key):
        """
        Performs the initial authentication on connect
        """
        return await self.call('auth', {"AuthKey": auth_key}, expect_body=False)

    @asynccontextmanager
    async def _connected(self):
        reader = ValueEvent()
        try:
            async with await anyio.connect_tcp(self.host, self.port) as sock:
                self._socket = sock
                await self.tg.spawn(self._reader, reader)
                reader = await reader.get()
                yield self
        except socket.error as e:
            raise SerfConnectionError(self.host, self.port) from e
        finally:
            if self._socket is not None:
                await self._socket.close()
                self._socket = None
            if reader is not None:
                await reader.cancel()
                reader = None

    @property
    def _counter(self):
        """
        Returns the current value of the iterator and increments it.
        """
        current = self._seq
        self._seq += 1
        return current

    def _decode_addr_key(self, obj_dict):
        """
        Callback function to handle the decoding of the 'Addr' field.

        Serf msgpack 'Addr' as an IPv6 address, and the data needs to be unpack
        using socket.inet_ntop().

        :param obj_dict: A dictionary containing the msgpack map.
        :return: A dictionary with the correct 'Addr' format.
        """
        key = b'Addr'
        if key in obj_dict:
            try:
                # Try to convert a packed IPv6 address.
                # Note: Call raises ValueError if address is actually IPv4.
                ip_addr = socket.inet_ntop(socket.AF_INET6, obj_dict[key])

                # Check if the address is an IPv4 mapped IPv6 address:
                # ie. ::ffff:xxx.xxx.xxx.xxx
                if ip_addr.startswith('::ffff:'):
                    ip_addr = ip_addr.lstrip('::ffff:')

                obj_dict[key] = ip_addr.encode('utf-8')
            except ValueError:
                # Try to convert a packed IPv4 address.
                ip_addr = socket.inet_ntop(socket.AF_INET, obj_dict[key])
                obj_dict[key] = ip_addr.encode('utf-8')

        return obj_dict




# Streaming frontend

class SerfStream:
    """
    Represents the data stream from a call to ``client.stream()``.
    """
    _it = None

    def __init__(self, client, stream):
        self.client = client
        self.stream = stream

    async def __aenter__(self):
        self._ctx = await self.stream.__aenter__()
        return self

    async def __aexit__(self, *tb):
        await self._ctx.__aexit__(*tb)
        self._ctx = None

    def __aiter__(self):
        self._it = self._ctx.__aiter__()
        return self

    async def __anext__(self):
        try:
            r = await self._it.__anext__()
        except StopAsyncIteration:
            self._it = None
            raise
        else:
            res = SerfEvent(self.client)
            for k,v in r.body.items():
                k = k.decode('UTF-8')
                if v is not None:
                    if k == "Payload":
                        v = self.client.codec.decode(v)
                    elif isinstance(v, bytes):
                        v = v.decode("utf-8")
                setattr(res, k.lower(), v)
            return res

    @property
    def head(self):
        return self._ctx.head
    @property
    def body(self):
        return self._ctx.body

    def cancel(self):
        if self._it is not None:
            return self._it.cancel()


class SerfQuery(SerfStream):
    def __init__(self, client, stream):
        super().__init__(client, stream)
        self.stream.send_stop = False

    async def __anext__(self):
        res = await super().__anext__()
        if res.type == "done":
            try:
                del self.client._conn._handlers[self.stream.seq]
            except AttributeError:
                pass
            raise StopAsyncIteration
        return res


class SerfEvent:
    """
    Encapsulates one event on a SerfStream.
    """
    id = None
    payload = None

    def __init__(self, client):
        self.client = client

    async def respond(self, payload):
        await self.client.respond(self.id, payload)
        if payload is not None:
            payload = self.client.codec.encode(payload)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, ",".join("%s:%s" %(str(k),repr(v)) for k,v in vars(self).items() if not k.startswith('_')))




import mock

from aioserf import EnvironmentConfig


@mock.patch('os.getenv')
def test_defaults(os_getenv):
    os_getenv.return_value = None

    env = EnvironmentConfig()

    assert env.host == 'localhost'
    assert env.port == 7373
    assert env.auth_key is None


@mock.patch('os.getenv')
def test_serf_rpc_addr(os_getenv):
    def fake_os_getenv(key, default=None):
        if key == 'SERF_RPC_ADDR':
            return 'serf.company.com:6464'

    os_getenv.side_effect = fake_os_getenv

    env = EnvironmentConfig()

    assert env.host == 'serf.company.com'
    assert env.port == 6464
    assert env.auth_key is None


@mock.patch('os.getenv')
def test_serf_rpc_auth(os_getenv):
    def fake_os_getenv(key, default=None):
        if key == 'SERF_RPC_AUTH':
            return 'secret'

    os_getenv.side_effect = fake_os_getenv

    env = EnvironmentConfig()

    assert env.host == 'localhost'
    assert env.port == 7373
    assert env.auth_key == 'secret'



import os


class EnvironmentConfig(object):
    """
    Reads environment variables that Serf understands and parses
    them into a easily consumable object.

    Currently supports reading the following environment variables:

     - SERF_RPC_AUTH
     - SERF_RPC_ADDR

    and sets attributes:

     - host
     - port
     - auth_key
    """

    def __init__(self):
        self.host = 'localhost'
        self.port = 7373
        self.auth_key = None

        rpc_addr = os.getenv('SERF_RPC_ADDR')
        if rpc_addr:
            self.host, self.port = rpc_addr.split(':')
            self.port = int(self.port)

        rpc_auth = os.getenv('SERF_RPC_AUTH')
        if rpc_auth:
            self.auth_key = rpc_auth



# utility code

import attr
import outcome

import anyio
from anyio.exceptions import CancelledError


@attr.s
class ValueEvent:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and tasks can wait for it to become True.

    """

    event = attr.ib(factory=anyio.create_event, init=False)
    value = attr.ib(default=None, init=False)

    async def set(self, value):
        """Set the result to return this value, and wake any waiting task.
        """
        self.value = outcome.Value(value)
        await self.event.set()

    async def set_error(self, exc):
        """Set the result to raise this exceptio, and wake any waiting task.
        """
        self.value = outcome.Error(exc)
        await self.event.set()

    def is_set(self):
        return self.value is not None

    def cancel(self):
        return self.set_error(CancelledError())

    async def get(self):
        """Block until the value is set.

        If it's already True, then this method is still a checkpoint, but
        otherwise returns immediately.

        """
        await self.event.wait()
        return self.value.unwrap()




#!/usr/bin/python3

import anyio
from aioserf import serf_client, UTF8Codec

import logging

async def main():
    async with serf_client(codec = UTF8Codec()) as client:
        async with client.stream('*') as stream:
            async for resp in stream:
                if resp.event == 'query':
                    await resp.respond('For %s, with %s' % (resp.name, resp.payload))

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    anyio.run(main)



import mock
import pytest
import re

from contextlib import closing
from aioserf import serf_client, SerfError


class TestAioSerfCommands(object):
    """
    Common commands for the library
    """
#    @pytest.mark.anyio
#    async def test_rpc_auth(self):
#        with mock.patch('aioserf.client.SerfConnection') as mock_serf_connection_class:
#            mock_serf_connection = mock.MagicMock()
#            mock_serf_connection_class.return_value = mock_serf_connection
#            async with serf_client(rpc_auth='secret') as serf:
#                assert serf._conn is not None
#            mock_serf_connection.auth.assert_called_once_with('secret')

    @pytest.mark.anyio
    async def test_has_a_default_host_and_port(self):
        async with serf_client() as serf:
            assert serf.host == 'localhost'
            assert serf.port == 7373

    @pytest.mark.anyio
    async def test_initialises_a_serf_connection_on_creation(self):
        async with serf_client() as serf:
            assert serf._conn is not None

    @pytest.mark.anyio
    async def test_sending_a_simple_event(self):
        async with serf_client() as serf:
            assert (await serf.event('foo', 'bar')).head == {b'Error': b'', b'Seq': 1}

    @pytest.mark.anyio
    async def test_sending_a_non_coalescing_event(self):
        async with serf_client() as serf:
            assert (await serf.event('foo', 'bar')).head == {b'Error': b'', b'Seq': 1}

    @pytest.mark.anyio
    async def test_event_payload_is_optional(self):
        async with serf_client() as serf:
            assert (await serf.event('foo')).head == {b'Error': b'', b'Seq': 1}
            assert (await serf.event('bar', coalesce=False)).head == \
                {b'Error': b'', b'Seq': 2}

    @pytest.mark.anyio
    async def test_force_leaving_of_a_node(self):
        async with serf_client() as serf:
            assert (await serf.force_leave('bad-node-name')).head == \
                {b'Error': b'', b'Seq': 1}

    @pytest.mark.anyio
    async def test_joining_a_non_existent_node(self):
        async with serf_client() as serf:
            address = '127.0.0.1:23000'
            with pytest.raises(SerfError) as e:
                join = await serf.join([address])
            assert 'dial tcp' in str(e.value)
            with pytest.raises(SerfError) as e:
                join = await serf.join([address])
            assert 'dial tcp' in str(e.value)

    @pytest.mark.anyio
    async def test_joining_an_existing_node_fails(self):
        async with serf_client() as serf:
            with pytest.raises(SerfError) as e:
                await serf.join(['127.0.0.1:7373'])
            join = e.value.args[0]
            assert join.head[b'Seq'] == 1
            assert b'EOF' in join.head[b'Error']
            assert join.body == {b'Num': 0}

    @pytest.mark.anyio
    async def test_providing_a_single_value_should_put_it_inside_a_list(self):
        async with serf_client() as serf:
            with pytest.raises(SerfError) as e:
                await serf.join('127.0.0.1:7373')
            join = e.value.args[0]
            assert join.head[b'Seq'] == 1
            assert b'EOF' in join.head[b'Error']
            assert join.body == {b'Num': 0}

    @pytest.mark.anyio
    async def test_member_list_is_not_empty(self):
        async with serf_client() as serf:
            members = await serf.members()
            assert len(members.body[b'Members']) > 0

    @pytest.mark.anyio
    async def test_member_filtering_name(self):
        async with serf_client() as serf:
            # Get current node name.
            members = await serf.members()
            name = members.body[b'Members'][0][b'Name']

            members = await serf.members(name=name)
            assert len(members.body[b'Members']) == 1

    @pytest.mark.anyio
    async def test_member_filtering_name_no_matches(self):
        async with serf_client() as serf:
            members = await serf.members(name="no_node_has_this_name")
            assert len(members.body[b'Members']) == 0

    @pytest.mark.anyio
    async def test_member_filtering_status_alive(self):
        async with serf_client() as serf:
            members = await serf.members(status="alive")
            assert len(members.body[b'Members']) > 0

    @pytest.mark.anyio
    async def test_member_filtering_status_no_matches(self):
        async with serf_client() as serf:
            members = await serf.members(status="invalid_status")
            assert len(members.body[b'Members']) == 0

    @pytest.mark.anyio
    async def test_member_filtering_tags(self):
        async with serf_client() as serf:
            members = await serf.members(tags={'foo': 'bar'})
            assert len(members.body[b'Members']) == 1

    @pytest.mark.anyio
    async def test_member_filtering_tags_regex(self):
        async with serf_client() as serf:
            members = await serf.members(tags={'foo': 'ba[rz]'})
            assert len(members.body[b'Members']) == 1

    @pytest.mark.anyio
    async def test_member_check_addr(self):
        async with serf_client() as serf:
            # regression test for issue 20
            members = await serf.members()
            ip_addr = members.body[b'Members'][0][b'Addr']

            assert re.match(b'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$', ip_addr) or b':' in ip_addr

    @pytest.mark.anyio
    async def test_stats_is_well_formed(self):
        async with serf_client() as serf:
            stats = await serf.stats()
            for key in [b'agent', b'runtime', b'serf', b'tags']:
                assert key in stats.body
                assert isinstance(stats.body[key], dict)

    @pytest.mark.anyio
    async def test_close(self):
        async with serf_client() as serf:
            assert serf._conn is not None
        assert serf._conn is None



from pkg_resources import get_distribution
from .client import AioSerf, serf_client
from .environment_config import EnvironmentConfig
from .exceptions import SerfError, SerfConnectionError
from . import codec

__all__ = ['AioSerf', 'serf_client', 'EnvironmentConfig', 'SerfError', 'SerfConnectionError']
__all__ += codec.__all__

for k in codec.__all__:
	globals()[k] = getattr(codec,k)



class SerfError(Exception):
    pass


class SerfConnectionError(SerfError):
    def __init__(self, host, port):
        self.host = host
        self.port = port
    def __str__(self):
        return "Error connecting to %s:%s" % (self.host, self.port)


class SerfClosedError(SerfError):
    pass



#
# async serf client
# (c) 2018 Matthias Urlichs

import anyio
from .connection import SerfConnection
from .stream import SerfStream, SerfQuery
from .codec import MsgPackCodec
from async_generator import asynccontextmanager

@asynccontextmanager
async def serf_client(**kw):
    async with anyio.create_task_group() as tg:
        client = AioSerf(tg, **kw)
        async with client._connected():
            yield client

class AioSerf(object):
    def __init__(self, tg, host='localhost', port=7373, rpc_auth=None,
            codec=None):
        self.tg = tg
        self.host = host
        self.port = port
        self.rpc_auth = rpc_auth
        if codec is None:
            codec = MsgPackCodec()
        self.codec = codec

    @asynccontextmanager
    async def _connected(self):
        self._conn = SerfConnection(self.tg, host=self.host, port=self.port)
        try:
            async with self._conn._connected():
                await self._conn.handshake()
                if self.rpc_auth:
                    await self._conn.auth(self.rpc_auth)
                yield self
        finally:
            if self._conn is not None:
                await self._conn.call("leave")
            self._conn = None

    def stream(self, event_types='*'):
        """
        Open an event stream.

        Possible event types:

        * ``*`` -- all events
        * ``user`` -- all user events
        * ``user:TYPE`` -- all user events of type TYPE
        * ``query`` -- all queries
        * ``query:TYPE`` -- all queries of type TYPE
        * ``member-join``
        * ``member-leave``

        This method returns a SerfStream object which affords an async
        context manager plus async iterator, which will give you SerfEvent
        objects, which you can use to return a reply (if triggered by a
        "query" event).

        Replies have a "respond" method which you may use to send
        responses to queries.

        Example Usage::

            >>> async with client.stream("user") as stream:
            >>>     async for event in stream:
            >>>         msg = await dispatch(event.type)(event.payload)
            >>>         await event.respond(msg)

        though you might want to process messages in a task group::

            >>> async def in_tg(event):
            >>>     msg = await dispatch(event.type)(event.payload)
            >>>     await event.respond(msg)
            >>> async with anyio.create_task_group() as tg:
            >>>     async with client.stream("query") as stream:
            >>>         async for event in stream:
            >>>             await tg.spawn(in_tg, event)

        """
        if isinstance(event_types, list):
            event_types = ','.join(event_types)
        res = self._conn.stream('stream', {'Type': event_types})
        return SerfStream(self, res)

    def query(self, name, payload=None, *, nodes=None, tags=None,
            request_ack=False, timeout=0):
        """
        Send a query, expect a stream of replies.
        """
        params = {'Name': name, 'RequestAck': request_ack}
        if payload is not None:
            params['Payload'] = self.codec.encode(payload)
        if nodes:
            params['FilterNodes'] = list(nodes)
        if tags:
            params['FilterTags'] = dict(tags)
        if timeout:
            params['Timeout'] = timeout

        res = self._conn.stream('query', params)
        return SerfQuery(self, res)

    def monitor(self, log_level='DEBUG'):
        res = self._conn.stream('monitor', {'LogLevel': log_level})
        return SerfStream(self, res)

    def respond(self, seq, payload):
        if seq is None:
            raise RuntimeError("You cannot respond to this message.")
        if payload is not None:
            payload = self.codec.encode(payload)
        return self._conn.call("respond", {'ID': seq, 'Payload': payload}, expect_body=False)


    def event(self, name, payload=None, coalesce=True):
        """
        Send an event to the cluster. Can take an optional payload as well,
        which will be sent in the form that it's provided.
        """
        if payload is not None:
            payload = self.codec.encode(payload)
        return self._conn.call(
            'event',
            {'Name': name, 'Payload': payload, 'Coalesce': coalesce},
            expect_body=False)

    def members(self, name=None, status=None, tags=None):
        """
        Lists members of a Serf cluster, optionally filtered by one or more
        filters:

        `name` is a string, supporting regex matching on node names.
        `status` is a string, supporting regex matching on node status.
        `tags` is a dict of tag names and values, supporting regex matching
        on values.
        """
        filters = {}

        if name is not None:
            filters['Name'] = name

        if status is not None:
            filters['Status'] = status

        if tags is not None:
            filters['Tags'] = tags

        if len(filters) == 0:
            return self._conn.call('members')
        else:
            return self._conn.call('members-filtered', filters)

    def tags(self, **tags):
        """Set this node's tags.

        Delete tags by setting its value to None.
        """
        deleted = []
        for k,v in list(tags.items()):
            if v is None:
                del tags[k]
                deleted.append(k)
        tags = {'Tags': tags}
        if deleted:
            tags['DeleteTags'] = deleted
        return self._conn.call('tags', tags)

    def force_leave(self, name):
        """
        Force a node to leave the cluster.
        """
        return self._conn.call(
            'force-leave',
            {"Node": name}, expect_body=False)

    def join(self, location):
        """
        Join another cluster by provided a list of ip:port locations.
        """
        if not isinstance(location, (list, tuple)):
            location = [location]
        return self._conn.call(
            'join',
            {"Existing": location, "Replay": False}, expect_body=2)

    def stats(self):
        """
        Obtain operator debugging information about the running Serf agent.
        """
        return self._conn.call('stats')




import pytest

from aioserf import result


class TestSerfResult(object):
    @pytest.fixture
    def head_and_body_result(self):
        return result.SerfResult(head=1, body=2)

    def test_initialises_to_none(self):
        r = result.SerfResult()
        assert r.head is None
        assert r.body is None

    def test_provides_a_pretty_printed_form_for_repl_use(self):
        r = result.SerfResult(head={"a": 1}, body=('foo', 'bar'))
        assert str(r) == \
            "SerfResult<head={'a': 1},body=('foo', 'bar')>"

    def test_can_convert_to_list_and_tuple(self, head_and_body_result):
        for structure in [list, tuple]:
            assert sorted(structure(head_and_body_result)) == [1, 2]



import anyio
import pytest
import re

from contextlib import closing
from aioserf import serf_client

NS=1000*1000*1000 # nanoseconds per second

class TestAioSerfStream(object):
    @pytest.mark.anyio
    async def test_sending_a_simple_event(self):
        async with serf_client() as serf:
            assert (await serf.event('foo', 'bar')).head == {b'Error': b'', b'Seq': 1}
            assert (await serf.event('bill', 'gates')).head == {b'Error': b'', b'Seq': 2}

    @pytest.mark.anyio
    async def test_stream(self):
        async with serf_client() as serf:
            async with serf.stream() as response:
                assert response.head == {b'Error': b'', b'Seq': 1}
                expected_data = sorted([
                    ['bill', 'gates'],
                    ['foo', 'bar'],
                ])
                all_responses = []
                async for resp in response:
                    all_responses.append(resp)
                    if len(all_responses) == 2:
                        break

            sorted_responses = sorted([
                [
                    res.name,
                    res.payload,
                ] for res in all_responses
            ])
            for i, res in enumerate(sorted_responses):
                expected = expected_data[i]
                assert res[0] == expected[0]
                assert res[1] == expected[1]


class TestAioSerfQuery(object):
    async def answer_query(self, serf, ev):
        async with serf.stream("query:foo") as s:
            await ev.set()
            async for r in s:
                assert r.payload == "baz"
                await r.respond("bar")

    async def ask_query(self, serf, ev):
        acks = 0
        reps = 0
        async with serf.query("foo", payload="baz", request_ack=True,
                timeout=1*NS) as q:
            async for r in q:
                if not hasattr(r,'type'):
                    break
                if r.type == "ack":
                    acks += 1
                elif r.type == "response":
                    reps += 1
                    assert r.payload == "bar"
                else:
                    assert False, r
        assert reps > 0
        assert acks > 0
        await ev.set()

    @pytest.mark.anyio
    async def test_query(self):
        async with anyio.create_task_group() as tg:
            async with serf_client() as serf1:
                async with serf_client() as serf2:
                    ev1 = anyio.create_event()
                    ev2 = anyio.create_event()
                    await tg.spawn(self.answer_query, serf2, ev1)
                    await ev1.wait()
                    await tg.spawn(self.ask_query, serf1, ev2)
                    await ev2.wait()



class TestAioSerfMonitor(object):
    @pytest.mark.anyio
    async def test_sending_a_simple_event(self):
        async with serf_client() as serf:
            assert (await serf.event('foo', 'bar')).head == {b'Error': b'', b'Seq': 1}
            assert (await serf.event('bill', 'gates')).head == {b'Error': b'', b'Seq': 2}

    @pytest.mark.anyio
    async def test_monitor(self):
        n = 0
        async with serf_client() as serf:
            async with serf.monitor() as response:
                async for resp in response:
                    print(resp)
                    n += 1
                    if n >= 3:
                        break




# Collection of basic codecs for the messages' payload

import msgpack

__all__ = ['UTF8Codec', 'MsgPackCodec']

class UTF8Codec:
    def encode(self, data):
        return data.encode("utf-8")
    def decode(self, data):
        return data.decode("utf-8")

class MsgPackCodec:
    def __init__(self, use_bin_type=True, use_list=False):
        self.use_bin_type = use_bin_type
        self.use_list = use_list
    def encode(self, data):
        return msgpack.packb(data, use_bin_type=self.use_bin_type)
    def decode(self, data):
        return msgpack.unpackb(data, raw=False, use_list=self.use_list)





#!/usr/bin/env python
import os
import sys

exec(open("aioserf/_version.py", encoding="utf-8").read())

try:
    from setuptools import setup
    from setuptools.command.test import test as TestCommand

    class PyTest(TestCommand):
        def finalize_options(self):
            TestCommand.finalize_options(self)
            self.test_args = []
            self.test_suite = True

        def run_tests(self):
            import pytest
            errno = pytest.main(self.test_args)
            sys.exit(errno)
except ImportError:
    from distutils.core import setup
    PyTest = lambda x: x

try:
    long_description = open(os.path.join(os.path.dirname(__file__),
                                         'README.rst')).read()
except:
    long_description = None

test_requires = [
    'pytest >= 2.5.2',
    'pytest-cov >= 2.3',
]

setup(
    name='aioserf',
    version=__version__,
    description='Python client for the Serf orchestration tool',
    long_description=long_description,
    url='https://github.com/smurfix/aioserf',
    author='Matthias Urlichs',
    author_email='matthias@urlichs.de',
    maintainer='Matthias Urlichs',
    maintainer_email='matthias@urlichs.de',
    keywords=['Serf', 'orchestration', 'service discovery', 'anyio'],
    license='MIT',
    packages=['aioserf'],
    install_requires=['msgpack >= 0.5.0', 'anyio', 'outcome', 'async_generator'],
    tests_require=test_requires,
    cmdclass={'test': PyTest},
)



import msgpack

class SerfResult(object):
    """
    Bounded result object for responses from a Serf agent.
    """

    def __init__(self, head=None, body=None):
        self.head, self.body = head, body

    def __iter__(self):
        yield self.head
        yield self.body

    def __repr__(self):
        return "%(class)s<head=%(h)s,body=%(b)s>" \
            % {'class': self.__class__.__name__,
               'h': self.head,
               'b': self.body}

